/*
 * Copyright (C) 1996 Be, Inc.  All Rights Reserved
 */
#include <netconfig.h>
#include <stdio.h>
#include <stdlib.h>
#include <socket.h>
#include <netdb.h>
#include <ether_driver.h>
#include <fcntl.h>
#include <string.h>
#include <unistd.h>
#include <Handler.h>
#include <TextControl.h>
#include <Window.h>
#include <Button.h>
#include <PopUpMenu.h>
#include <MenuField.h>
#include <MenuBar.h>
#include <MenuItem.h>
#include <TextView.h>

/*
 * NB: for many ISA cards, you may only need to change the DRIVER_NAME
 * and TITLE to make this add-on work for your card.  For PCI, you will
 * also want to remove all of the ISA parameter code.
 */
static const char DRIVER_NAME[] = "/dev/ether";			/* our device-driver */
static const char TITLE[] = "3Com Etherlink II (ISA)";	/* our pretty name */

/*
 * Default ISA values for this card
 */
#define DEFAULT_PORT 0x300
#define DEFAULT_IRQ  5
#define DEFAULT_MEM  0xc8000

/*
 * Stuff that gets passed down to the card during initialization
 */
typedef struct ether_init_params {
	short port;
	short irq;
	unsigned long mem;
} ether_init_params_t;

typedef unsigned long ip_address_t;	/* an IP address */

/*
 * Message codes
 */
enum control_codes {
	IPHIT = 0x1000,
	NMHIT,
	OKHIT,
	CANCELHIT
};

/*
 *  Hard-code some UI characteristics
 */
#define BUTTON_HEIGHT 18

static const rgb_color BACKGROUND_COLOR = { 216, 216, 216, 0 };
static const rgb_color WHITE_COLOR = { 255, 255, 255, 0 };
static const rgb_color BLACK_COLOR = { 0, 0, 0, 0 };


static BRect WINRECT(		140,	140,		380,		370);

static BRect TITLEFRAME(	10,		10,		200,		25);
static BRect IPFRAME(		10,		30,		200,		55);
static BRect NMFRAME(		10,		70,		200,		85);

static BRect PORTFRAME(		10,		100,	200,		115);

static BRect IRQFRAME(		10,		130,	200,		145);

static BRect MEMFRAME(		10,		160,	200,		175);
static BRect CANCELFRAME(	10,		190,	90,			190 + BUTTON_HEIGHT);
static BRect OKFRAME(		120,	190,	200,		190 + BUTTON_HEIGHT);

/*
 * Our window class
 */
class TWindow : public BWindow {
	sem_id sem;
	int status;
public:
	TWindow(BRect frame,
			const char *title, 
			window_type type,
			ulong flags,
			ulong workspace = B_CURRENT_WORKSPACE)
	: BWindow(frame, title, type,  flags, workspace) {
		sem = create_sem(0, "tulip window");
		status = 0;
	}
	~TWindow(void) {
		delete_sem(sem);
	}
	int Wait(void) {
		acquire_sem(sem);
		return (status);
	}
	void MessageReceived(BMessage *msg) {
		switch (msg->what) {
		case IPHIT:
			break;
		case NMHIT:
			break;
		case OKHIT:
			release_sem(sem);
			break;
		case CANCELHIT:
			status = -1;
			release_sem(sem);
			break;
		}
	}
	bool FilterKeyDown(ulong *key, BView **targetView) {
		if (*key == B_ESCAPE) {
			status = -1;
			release_sem(sem);
		}
		return (inherited::FilterKeyDown(key, targetView));
	}
};

/*
 * compute a text rect, given a frame
 */
static BRect 
textrect(BRect frame)
{
 	frame.InsetBy(2, 2);
	frame.OffsetTo(0, 0);
	return (frame);
}

/*
 * Initialize the text control
 */
static void
textcontrol_init(
				 BTextControl *tc
				 )
{
	BTextView *tv;

	tv = (BTextView *)tc->ChildAt(0);
	tv->SetViewColor(WHITE_COLOR);
	tv->SetHighColor(BLACK_COLOR);
	tv->SetLowColor(WHITE_COLOR);
}

/*
 * Inititialize the title view 
 */
static void
title_init(
		  BTextView *view,
		  const char *text
		  )
{
	
	view->MakeEditable(FALSE);
	view->MakeSelectable(FALSE);
	view->SetViewColor(BACKGROUND_COLOR);
	view->SetHighColor(BLACK_COLOR);
	view->SetLowColor(BACKGROUND_COLOR);
	view->SetText(text);
}

/*
 * Parse the configuration data to determine the ISA settings
 */
static void
get_card_stuff(
			   unsigned nprivate,
			   nc_private_t *private_items,
			   int *port,
			   int *irq,
			   int *mem
			   )
{
	int i;

	*port = DEFAULT_PORT;
	*irq = DEFAULT_IRQ;
	*mem = DEFAULT_MEM;
	for (i = 0; i < nprivate; i++) {
		if (strcmp(private_items[i].key, "DEVICEPORT") == 0) {
			sscanf(private_items[i].value, "%x", port);
		} else if (strcmp(private_items[i].key, "DEVICEIRQ") == 0) {
			sscanf(private_items[i].value, "%d", irq);
		} else if (strcmp(private_items[i].key, "DEVICEMEM") == 0) {
			sscanf(private_items[i].value, "%x", mem);
		}
	}
}

/*
 * Convert the ISA settings into configuration data format
 */
static void
set_card_stuff(
			   unsigned *nprivatep,
			   nc_private_t *private_items,
			   int port,
			   int irq,
			   int mem
			   )
{
	int i;
	int nprivate = *nprivatep;
	enum {
		HAVE_DEVICEPORT = 0x01,
		HAVE_DEVICEIRQ = 0x02,
		HAVE_DEVICEMEM = 0x04,
	};
	int have;

	have = 0;
	for (i = 0; i < nprivate; i++) {
#define CHECK(x) if (strcmp(private_items[i].key, #x) == 0) have |= HAVE_##x

		CHECK(DEVICEIRQ);
		CHECK(DEVICEPORT);
		CHECK(DEVICEMEM);
#undef CHECK
	}

#define DOIT(x) \
	if (!(have & HAVE_##x)) strcpy(private_items[nprivate++].key, #x)

	DOIT(DEVICEPORT);
	DOIT(DEVICEIRQ);
	DOIT(DEVICEMEM);
#undef DOIT
	*nprivatep = nprivate;

	for (i = 0; i < nprivate; i++) {
		if (strcmp(private_items[i].key, "DEVICEPORT") == 0) {
			sprintf(private_items[i].value, "%x", port);
		} else if (strcmp(private_items[i].key, "DEVICEIRQ") == 0) {
			sprintf(private_items[i].value, "%d", irq);
		} else if (strcmp(private_items[i].key, "DEVICEMEM") == 0) {
			sprintf(private_items[i].value, "%x", mem);
		}
	}
}

/*
 * Set the font for the menu field
 */
static void
set_font(
		 BPopUpMenu *menu,
		 BMenuField *field
		 )
{
	field->SetFontName("Erich");
	field->MenuBar()->SetFontName("Erich");
	menu->SetFontName("Erich");
}

/*
 * Here are the entry points: the pragma causes these entry points to 
 * exported from the add-on.
 */
#pragma export on

/*
 * configure: put up a window to let the user set the 
 * IP address, netmask and ISA settings for this card.
 *
 * Returns a non-negative number if the new settings should be accepted, or
 * a negative number indicating the user canceled the operation.
 *
 */
extern "C" long
configure(
		  ip_address_t *ipaddressp,
		  ip_address_t *netmaskp,
		  char *driver_name,
		  unsigned *nprivate,
		  nc_private_t *private_items
		  )
{
	TWindow *win;
	BTextControl *ipedit;
	BTextControl *nmedit;
	BButton *okbut;
	BButton *cancelbut;
	BView *background;
	struct in_addr ip;
	struct in_addr nm;
	char ascii_ip[16];
	char ascii_nm[16];
	BMenuField *portmbar;
	BMenuField *irqmbar;
	BMenuField *memmbar;
	BPopUpMenu *portmenu;
	BPopUpMenu *irqmenu;
	BPopUpMenu *memmenu;
	BMenuItem *mitem;
	rgb_color hi;
	rgb_color lo;
	BTextView *title;
	int port;
	int irq;
	int mem;
	int status;
	
	get_card_stuff(*nprivate, private_items, &port, &irq, &mem);

	ip.s_addr = *ipaddressp;
	nm.s_addr = *netmaskp;
	strcpy(ascii_ip, inet_ntoa(ip));
	strcpy(ascii_nm, inet_ntoa(nm));
	win = new TWindow(WINRECT, TITLE, B_MODAL_WINDOW, 
					  B_NOT_RESIZABLE | B_NOT_ZOOMABLE | B_NOT_CLOSABLE);


	title = new BTextView(TITLEFRAME, "title", textrect(TITLEFRAME), 
						  B_FOLLOW_NONE, B_WILL_DRAW);

	ipedit = new BTextControl(IPFRAME, "", "IP Address", 
							  ascii_ip, new BMessage(IPHIT));
	ipedit->SetFontName("Erich");
	nmedit = new BTextControl(NMFRAME, "", "Netmask", ascii_nm, 
							  new BMessage(NMHIT));
	nmedit->SetFontName("Erich");
	okbut = new BButton(OKFRAME, "", "Set", new BMessage(OKHIT));
	cancelbut = new BButton(CANCELFRAME, "", "Cancel", new BMessage(CANCELHIT));



	portmenu = new BPopUpMenu("port");
	portmbar = new BMenuField(PORTFRAME, "", "Port", portmenu);
	set_font(portmenu, portmbar);

#define ADD_PORT_ITEM(N) \
	portmenu->AddItem(mitem = new BMenuItem(#N,  NULL)); \
	if (0x##N == port) mitem->SetMarked(TRUE);
		
	
	ADD_PORT_ITEM(250);
	ADD_PORT_ITEM(280);
	ADD_PORT_ITEM(2A0);
#if 0
	ADD_PORT_ITEM(2E0); /* reserved for com4 */
#endif
	ADD_PORT_ITEM(300); 
	ADD_PORT_ITEM(310);
	ADD_PORT_ITEM(330);
	ADD_PORT_ITEM(350);


	irqmenu = new BPopUpMenu("irq");
	irqmbar = new BMenuField(IRQFRAME, "", "IRQ", irqmenu);
	set_font(irqmenu, irqmbar);

#define ADD_IRQ_ITEM(N) \
	irqmenu->AddItem(mitem = new BMenuItem(#N,  NULL)); \
	if (N == irq) mitem->SetMarked(TRUE);
	
	ADD_IRQ_ITEM(2);
#if 0
	ADD_IRQ_ITEM(3);	/* reserved for com2 */
	ADD_IRQ_ITEM(4);	/* reserved for com1 */
#endif
	ADD_IRQ_ITEM(5);

	memmenu = new BPopUpMenu("mem");
	memmbar = new BMenuField(MEMFRAME, "", "Address", memmenu);
	set_font(memmenu, memmbar);

#define ADD_MEM_ITEM(N) \
	memmenu->AddItem(mitem = new BMenuItem(#N,  NULL)); \
	if (0x##N == mem) mitem->SetMarked(TRUE);
	
	ADD_MEM_ITEM(c8000);
	ADD_MEM_ITEM(cc000);
	ADD_MEM_ITEM(d8000);
	ADD_MEM_ITEM(dc000);


	background = new BView(win->Bounds(), "", B_FOLLOW_ALL, B_WILL_DRAW);

	win->Lock();
	win->AddChild(background);
	background->SetViewColor(BACKGROUND_COLOR);

	background->AddChild(title);
	background->AddChild(ipedit);
	background->AddChild(nmedit);
	background->AddChild(portmbar);
	background->AddChild(irqmbar);
	background->AddChild(memmbar);
	background->AddChild(cancelbut);
	background->AddChild(okbut);
	okbut->MakeDefault(TRUE);

	title_init(title, TITLE);
	textcontrol_init(ipedit);
	textcontrol_init(nmedit);
	
	ipedit->MakeFocus(TRUE);
	win->Unlock();

	win->Show();
	status = win->Wait();
	if (status >= 0) {
		int i;

		for (i = 0; i < portmenu->CountItems(); i++) {
			mitem = portmenu->ItemAt(i);
			if (mitem->IsMarked()) {
				sscanf(mitem->Label(), "%x", &port);
				break;
			}
		}

		for (i = 0; i < irqmenu->CountItems(); i++) {
			mitem = irqmenu->ItemAt(i);
			if (mitem->IsMarked()) {
				sscanf(mitem->Label(), "%d", &irq);
				break;
			}
		}

		for (i = 0; i < memmenu->CountItems(); i++) {
			mitem = memmenu->ItemAt(i);
			if (mitem->IsMarked()) {
				sscanf(mitem->Label(), "%x", &mem);
				break;
			}
		}
		set_card_stuff(nprivate, private_items, port, irq, mem);
		*ipaddressp = inet_addr(ipedit->Text());
		*netmaskp = inet_addr(nmedit->Text());
		strcpy(driver_name, DRIVER_NAME);
	}

	win->Quit();

	return (status);
}

/*
 * Open the driver, and initialize it using the configuration data 
 * ISA settings. Return the file descriptor associated with the
 * device-driver, or -1 upon error.
 *
 */
extern "C" int
driver_open(const char *device)
{
	char devicelink[NC_MAXVALLEN+1];
	char deviceirq[NC_MAXVALLEN+1];
	char deviceport[NC_MAXVALLEN+1];
	char devicemem[NC_MAXVALLEN+1];
	int port;
	int irq;
	unsigned long mem;
	ether_init_params_t params;
	int fd;

	if (!netconfig_find(device, "DEVICELINK", devicelink)) {
		return (-1);
	}
	if (!netconfig_find(device, "DEVICEIRQ", deviceirq)) {
		irq = DEFAULT_IRQ;
	} else {
		irq = atoi(deviceirq);
	}
	if (!netconfig_find(device, "DEVICEPORT", deviceport)) {
		port = DEFAULT_PORT;
	} else {
		sscanf(deviceport, "%x", &port);
	}
	if (!netconfig_find(device, "DEVICEMEM", devicemem)) {
		mem = DEFAULT_MEM;
	} else {
		sscanf(devicemem, "%x", &mem);
	}
	params.port = port;
	params.irq = irq;
	params.mem = mem;
	fd = open(devicelink, O_RDWR, 0);
	if (fd < 0) {
		return (-1);
	}
	if (ioctl(fd, ETHER_INIT, (char *)&params, 0) < 0) {
		close(fd);
		return (-1);
	}
	return (fd);
}

/*
 * Return the pretty name for this driver
 */
extern "C" void
get_pretty_name(char *pretty_name)
{
	strcpy(pretty_name, TITLE);
}

#pragma export reset
